
import java.applet.Applet;
import java.awt.Button;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Label;
import java.awt.List;
import java.awt.Panel;
import java.awt.ScrollPane;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

/**
 * The Huffman algorithm demo applet. Typical applet: almost everything is done
 * in init(). Using the Java 1.1 event model, we use inner classes to react to 
 * the buttons; mostly, some support method are called.<br>
 * This code is straight-forward - run the applet, and you see what the code 
 * does in detail.
 * <p>
 * (c) R.Reichert & C.Oehler 1998, minor revision C. Sigg, 2004
 */
public  class HuffApplet extends Applet  {      

    public Color inputBarColor    = Color.darkGray;
    public Color outputBarColor   = Color.blue;
    public Color borderPanelColor = Color.black;
    public Color labelColor       = Color.black;
    public Color circleColor      = Color.blue; 

    //
    // SimpleBar is used to display very simple bar charts;
    // in particular, to give a graphical impression of the compression
    // rate. The constructor parameter factor gives the percentage of
    // the width of this component that is to be filled with the bar;
    // the whole height is used. This makes it easy to line up some
    // SimpleBar's vertically, and to set their width accordingly.
    // 
    private class SimpleBar extends Component {
	private double factor;
	private Color barColor;
	
	public SimpleBar (Color barColor, double factor) {
	    this.barColor = barColor;
	    this.factor = factor;
	} /* SimpleBar */
	
	public SimpleBar () {
	    this (Color.black, 0.0);
	} /* SimpleBar */
	
	public void setFactor (double factor) {
	    this.factor = factor;
	    repaint();
	} /* SimpleBar */

	public void setColor (Color barColor) {
	    this.barColor = barColor;
	    repaint();
	} /* setColor */

	public void paint (Graphics g) {
	    Dimension size = getSize();
	    
	    g.setColor (getBackground());
	    g.fillRect (0, 0, size.width, size.height);

	    if (factor > 0.0) {
		g.setColor (barColor);		
		g.fillRect (0, 0, (int) (size.width*factor), size.height);
	    }
	} /* paint */
	
    } /* SimpleBar */

    // 
    // We wanted a panel that has a border drawn around it; unfortunately,
    // such a class will only be available in Java 1.2 (Swing). So this is
    // just a quick&dirty hack.
    //
    private class SimpleBorderPanel extends Panel {
	private Color lineColor;
	private String label;

	public SimpleBorderPanel (Color color, String label) {
	    lineColor = color;
	    this.label = new String (" "+label+" ");
	} /* SimpleBorderPanel */

	public Insets getInsets() {
	    return new Insets (20, 10, 10, 10); 
	} /* getInsets */

	public void paint (Graphics g) {
	    super.paint (g);

	    Dimension s = getSize();
	    int strHeight= g.getFontMetrics().getHeight();
	    int strWidth = g.getFontMetrics().stringWidth (label);

	    g.setColor (lineColor);
	    g.drawRect (2, 2+strHeight/2, s.width-4, s.height-4-strHeight/2);
	    g.setColor (getBackground());
	    g.drawLine ((s.width-strWidth)/2, 2+strHeight/2, 
			(s.width+strWidth)/2, 2+strHeight/2); 
	    g.setColor (lineColor);
	    g.drawString (label, (s.width-strWidth)/2, 2+strHeight); 
	} /* paint */

    } /* SimpleBorderPanel */

    public Huffman huff;           // the Huffman encoder

    public Panel borderPanel1;      // top most panel:
    public TextField plainText;     //   input field
    public Button btCompress;       //   compress-at-once button
    public Button btStepwise;       //   compress-stepwise button
    public Button btClear;          //   clear all button

    public Panel borderPanel2;      // the Huffman panel: 
    public Label probLabel;         //   probabilities-list label
    public List probList;           //   probabilities-list
    public Label huffLabel;         //   huff component label
    public HuffComponent hc;        //   huffmann tree component
    public Label codesLabel;        //   codes-list label
    public List codeList;           //   codes list
    public Button btNextStep;       //   next-step button
    public Button btFinish;         //   finish-compression button

    public Panel borderPanel3;      // the output panel: 
    public TextArea codedText;      //   binary huffmann output

    public Panel borderPanel4;      // the statistics panel:
    public Label labelIBar;         //   input-length bar label
    public SimpleBar inputBar;      //   input-length bar
    public Label labelOBar;         //   output-length bar label   
    public SimpleBar outputBar;     //   output-length bar label
    public Label lbAVGLength;       //   average-length label
    public Label avgLength;         //   average-length 
   

    public void init () {
	this.setLayout (new GridBagLayout());

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	borderPanel1 = new SimpleBorderPanel (borderPanelColor, "Your Input");
	{
	    borderPanel1.setLayout (new GridBagLayout());

	    addConstrained (borderPanel1, plainText = new TextField(), 
			    0 /* gridx */, 0 /* gridy */,
			    3 /* gridwidth */, 1 /* gridheight */, 
			    GridBagConstraints.HORIZONTAL,
			    GridBagConstraints.NORTHWEST,
			    1.0 /* weightx */, 0.0 /* weighty */, 
			    2, 2, 2, 2 /* insets */);

	    btClear = new Button ("Clear");
	    btClear.addActionListener (new ActionListener() {
		public void actionPerformed (ActionEvent e) {  
		    plainText.setText (""); 
		    clear (); 
		} /* actionPerformed */
	    });	    
	    addConstrained (borderPanel1, btClear, 
			    4 /* gridx */, 0 /* gridy */);

	    btCompress = new Button ("Compress");
	    btCompress.addActionListener (new ActionListener() {
		public void actionPerformed (ActionEvent e) {  
		    startCompression();
		    if (huff != null) {
			huff.compress();
			endCompression();
		    }
		} /* actionPerformed */
	    });	    
	    addConstrained (borderPanel1, btCompress, 
			    0 /* gridx */, 1 /* gridy */);

	    btStepwise = new Button ("Stepwise Compression");
	    btStepwise.addActionListener (new ActionListener() {
		public void actionPerformed (ActionEvent e) {  
		    startCompression();
		    if (huff != null) {
			hc.update();
			makeCodeList();
			btNextStep.setEnabled (true);
			btFinish.setEnabled (true);
		    }
		} /* actionPerformed */
	    });	    
	    addConstrained (borderPanel1, btStepwise, 
			    1 /* gridx */, 1 /* gridy */);
	}
	addConstrained (this, borderPanel1, 
			0 /* gridx */, 0 /* gridy */,
			1 /* gridwidth */, 1 /* gridheight */, 
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.NORTHWEST,
			1.0 /* weightx */, 0.0 /* weighty */, 
			2, 2, 2, 2 /* insets */);

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 	borderPanel2 = new SimpleBorderPanel (borderPanelColor, "Inside Huffman");
 	{
	    borderPanel2.setLayout (new GridBagLayout());

	    Panel subPanel = new Panel(); 
	    subPanel.setLayout (new GridBagLayout());
	    // for some reason unknown to me, lists disturb the layout
	    // of the applet (eg, they insert scrollbars without resizing
	    // themselves, thereby overpainting the object next to them)
	    // unless they are inserted in a sub-panel of their own. 
	    // Hence this is just what I do here.
	    // Out-commented below is the original code:
//
// 	    addConstrained (borderPanel2, probLabel = new Label("Probabilities"), 
// 			    0 /* gridx */, 0 /* gridy */);
// 	    addConstrained (borderPanel2, probList = new List (0, false), 
// 			    0 /* gridx */, 1 /* gridy */,
// 			    1 /* gridwidth */, 4 /* gridheight */, 
// 			    GridBagConstraints.VERTICAL,
// 			    GridBagConstraints.NORTHWEST,
// 			    0.0 /* weightx */, 1.0 /* weighty */, 
// 			    2, 2, 2, 2 /* insets */);

	    addConstrained (subPanel, probLabel = new Label("Probabilities"), 
			    0 /* gridx */, 0 /* gridy */);

	    addConstrained (subPanel, probList = new List (0, false), 
 			    0 /* gridx */, 1 /* gridy */,
 			    1 /* gridwidth */, 1 /* gridheight */, 
 			    GridBagConstraints.VERTICAL,
 			    GridBagConstraints.NORTHWEST,
 			    0.0 /* weightx */, 1.0 /* weighty */, 
 			    2, 2, 2, 2 /* insets */);
	  
	    addConstrained (borderPanel2, subPanel, 
 			    0 /* gridx */, 0 /* gridy */,
 			    1 /* gridwidth */, 5 /* gridheight */, 
 			    GridBagConstraints.VERTICAL,
 			    GridBagConstraints.NORTHWEST,
 			    0.0 /* weightx */, 1.0 /* weighty */, 
 			    0, 0, 0, 0 /* insets */);

	    addConstrained (borderPanel2, huffLabel = new Label("Huffman Tree"), 
			    1 /* gridx */, 0 /* gridy */);

	    ScrollPane sp = new ScrollPane (ScrollPane.SCROLLBARS_ALWAYS);
	    sp.add (hc = new HuffComponent (labelColor, circleColor));
	    addConstrained (borderPanel2, sp, 
			    1 /* gridx */, 1 /* gridy */,
			    5 /* gridwidth */, 4 /* gridheight */, 
			    GridBagConstraints.BOTH,
			    GridBagConstraints.CENTER,
			    1.0 /* weightx */, 1.0 /* weighty */, 
			    2, 2, 2, 2 /* insets */);

	    subPanel = new Panel(); 
	    subPanel.setLayout (new GridBagLayout());	    
// 	    addConstrained (borderPanel2, codesLabel = new Label("Codes"), 
// 			    3 /* gridx */, 0 /* gridy */);
// 	    addConstrained (borderPanel2, codeList = new List (0, false), 
// 			    3 /* gridx */, 1 /* gridy */,
// 			    1 /* gridwidth */, 4 /* gridheight */, 
// 			    GridBagConstraints.VERTICAL,
// 			    GridBagConstraints.NORTHWEST,
// 			    0.0 /* weightx */, 1.0 /* weighty */, 
// 			    2, 2, 2, 2 /* insets */);
 	    addConstrained (subPanel, codesLabel = new Label("Codes"), 
 			    0 /* gridx */, 0 /* gridy */);
	    addConstrained (subPanel, codeList = new List (0, false), 
			    0 /* gridx */, 1 /* gridy */,
			    1 /* gridwidth */, 1 /* gridheight */, 
			    GridBagConstraints.VERTICAL,
			    GridBagConstraints.NORTHWEST,
			    0.0 /* weightx */, 1.0 /* weighty */, 
			    2, 2, 2, 2 /* insets */);
	    addConstrained (borderPanel2, subPanel, 
 			    6 /* gridx */, 0 /* gridy */,
 			    1 /* gridwidth */, 5 /* gridheight */, 
 			    GridBagConstraints.VERTICAL,
 			    GridBagConstraints.NORTHWEST,
 			    0.0 /* weightx */, 1.0 /* weighty */, 
 			    0, 0, 0, 0 /* insets */);

	    btNextStep = new Button ("Next Step");
	    btNextStep.setEnabled (false);
	    btNextStep.addActionListener (new ActionListener() {
		public void actionPerformed (ActionEvent e) {  
		    if (huff.oneStep()) 
			endCompression();
		    else {
			hc.update();
			makeCodeList(); 
		    }
		} /* actionPerformed */
	    });	    
	    addConstrained (borderPanel2, btNextStep, 
			    1 /* gridx */, 5 /* gridy */);

	    btFinish = new Button ("Finish Up");
	    btFinish.setEnabled (false);
	    btFinish.addActionListener (new ActionListener() {
		public void actionPerformed (ActionEvent e) {  
		    huff.compress();
		    endCompression();
		} /* actionPerformed */
	    });	    
	    addConstrained (borderPanel2, btFinish, 
			    2 /* gridx */, 5 /* gridy */);
	}
	addConstrained (this, borderPanel2, 
			0 /* gridx */, 1 /* gridy */,
			1 /* gridwidth */, 4 /* gridheight */, 
			GridBagConstraints.BOTH,
			GridBagConstraints.CENTER,
			1.0 /* weightx */, 1.0 /* weighty */, 
			2, 2, 2, 2 /* insets */);

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	borderPanel3 = new SimpleBorderPanel (borderPanelColor, "Output");
	{
	    borderPanel3.setLayout (new GridBagLayout());
	    codedText = new TextArea ("", 1, 20, TextArea.SCROLLBARS_HORIZONTAL_ONLY);
	    codedText.setEditable (false);
	    codedText.setCursor (new Cursor (Cursor.DEFAULT_CURSOR));
	    addConstrained (borderPanel3, codedText, 
			    0 /* gridx */, 0 /* gridy */,
			    1 /* gridwidth */, 1 /* gridheight */, 
			    GridBagConstraints.HORIZONTAL,
			    GridBagConstraints.NORTHWEST,
			    1.0 /* weightx */, 0.0 /* weighty */, 
			    2, 2, 2, 2 /* insets */);
	}
	addConstrained (this, borderPanel3, 
			0 /* gridx */, 5 /* gridy */,
			1 /* gridwidth */, 1 /* gridheight */, 
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.NORTHWEST,
			1.0 /* weightx */, 0.0 /* weighty */, 
			2, 2, 2, 2 /* insets */);

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	borderPanel4 = new SimpleBorderPanel (borderPanelColor, "Statistics");
	{
	    borderPanel4.setLayout (new GridBagLayout());

	    addConstrained (borderPanel4, labelIBar = new Label("Length of Input (ASCII): "), 
			    0 /* gridx */, 0 /* gridy */);
	    addConstrained (borderPanel4, inputBar = new SimpleBar (inputBarColor, 0.0),
			    1 /* gridx */, 0 /* gridy */,
			    1 /* gridwidth */, 1 /* gridheight */, 
			    GridBagConstraints.BOTH,
			    GridBagConstraints.NORTHWEST,
			    1.0 /* weightx */, 0.0 /* weighty */, 
			    2, 2, 2, 2 /* insets */);

	    addConstrained (borderPanel4, labelOBar = new Label("Length of Output (Huffman): "), 
			    0 /* gridx */, 1 /* gridy */);
	    addConstrained (borderPanel4, outputBar = new SimpleBar (outputBarColor, 0.0),
			    1 /* gridx */, 1 /* gridy */,
			    1 /* gridwidth */, 1 /* gridheight */, 
			    GridBagConstraints.BOTH,
			    GridBagConstraints.NORTHWEST,
			    1.0 /* weightx */, 0.0 /* weighty */, 
			    2, 2, 2, 2 /* insets */);	  
	    
	    addConstrained (borderPanel4, lbAVGLength = new Label("Average Codeword Length: "), 
			    0 /* gridx */, 2 /* gridy */);
	    addConstrained (borderPanel4, avgLength = new Label(""), 
			    1 /* gridx */, 2 /* gridy */);
	}
	addConstrained (this, borderPanel4, 
			    0 /* gridx */, 6 /* gridy */,
			    1 /* gridwidth */, 1 /* gridheight */, 
			    GridBagConstraints.HORIZONTAL,
			    GridBagConstraints.NORTHWEST,
			    1.0 /* weightx */, 0.0 /* weighty */, 
			    2, 2, 2, 2 /* insets */);
    } /* init */

    private void addConstrained 
        (Container container, Component component, 
	 int grid_x, int grid_y, int grid_width, int grid_height,
	 int fill, int anchor, double weight_x, double weight_y,
	 int top, int left, int bottom, int right)
    {
        GridBagConstraints c = new GridBagConstraints();
        c.gridx = grid_x; c.gridy = grid_y;
        c.gridwidth = grid_width; c.gridheight = grid_height;
        c.fill = fill; c.anchor = anchor;
        c.weightx = weight_x; c.weighty = weight_y;
        if (top+bottom+left+right > 0)
            c.insets = new Insets(top, left, bottom, right);
        
	container.add (component, c);
    } /* addConstrained */
    	
    private void addConstrained 
        (Container container, Component component, 
	 int grid_x, int grid_y)
    {
	addConstrained (container, component, grid_x, grid_y, 
			1 /* gridwidth */, 1 /* gridheight */, 
			GridBagConstraints.NONE, GridBagConstraints.NORTHWEST,
			0.0 /* weightx */, 0.0 /* weighty */, 
			2, 2, 2, 2 /* insets */);
    } /* addConstrained */
    
    public void makeProbList () {
	TreeNode[] nodes = huff.getNodesArray();

	probList.removeAll();
	if (nodes != null) 
	    for (int i = 0; i < nodes.length; i++)
		probList.add("'"+nodes[i].theChar + "': " + 
				   HuffComponent.formatFloat (nodes[i].prob));
    } /* makeProbList */
	    
    public void makeCodeList () {
	TreeNode[] nodes = huff.getNodesArray();

	codeList.removeAll();
	if (nodes != null) 
	    for (int i = 0; i < nodes.length; i++)
		codeList.add("'"+nodes[i].theChar + "': " + nodes[i].code);
    } /* makeCodeList */
	    
    public void startCompression() {
	String input = plainText.getText ();
	clear();
	if (input.length() > 0) {
	    huff = new Huffman (plainText.getText ());   
	    makeProbList();
	    hc.setHuffman (huff); 
	    inputBar.setFactor (0.0);
	    outputBar.setFactor (0.0);
	}
    } /* startCompression */

    public void endCompression() {
	// pre: huff != null
	hc.update();
	makeCodeList();	  
	
	btNextStep.setEnabled (false);
	btFinish.setEnabled (false);

	codedText.setText (huff.output ()); 
	inputBar.setFactor (1.0);
	outputBar.setFactor (huff.getAvgLength()/8.0);
	avgLength.setText (String.valueOf (huff.getAvgLength ()));

	borderPanel4.doLayout();
    } /* endCompression */

    public void clear () {
	// empty lists
	probList.removeAll(); 
	codeList.removeAll(); 
	// clear text fields 
	codedText.setText ("");
	avgLength.setText ("");
	inputBar.setFactor (0.0);
	outputBar.setFactor (0.0);
	// clear huffman tree
	hc.setHuffman (huff = null); 
	hc.update();
    } /* clear */

    public static void main (String[] args) {
	Frame frame = new Frame ("Huffman Demo Applet");
	Applet a = new HuffApplet();
	frame.add (a, "Center");
	a.init();
	frame.setSize (800, 700);
	frame.setVisible(true);
	frame.addWindowListener (new WindowAdapter() {
	    public void windowClosing (WindowEvent e) { System.exit (0); }
	});
    } /* main */

} /* HuffApplet */


